#ifndef MpsAnalysis_HH
#define MpsAnalysis_HH

#include <cassert>

#include <vector>
#include <iterator>
#include <map>
#include <iostream>
#include <fstream>
#include <sstream>  // Required for stringstreams
#include <string> 

#include "RcdUserRO.hh"
#include "SubAccessor.hh"

#include "TH1F.h"
#include "TFile.h"
#include "TH2F.h"
#include "TH3F.h"
#include "TH1I.h"
#include "TF1.h"
#include "TProfile.h"
#include "TObject.h"

#include "MpsSensor1FullTimeStamps.hh"

//Handy conversion for dealing with sensor numbers in ROOT file titles etc.
std::string IntToStr(int number) {
	std::ostringstream oss;

	// Works just like cout
	oss<< number;

	// Return the underlying string
	return oss.str();
}

class MpsAnalysis : public RcdUserRO {

public:
	enum {
		defaultMpsAnalysisBits=0xffff
	};

	MpsAnalysis(unsigned wombat, int runNumber, int bins = 1000, int low = 0,
			int high = 2000, int trackingSensor = 0);

	virtual ~MpsAnalysis();

	bool record(const RcdRecord& r);

	void handleBunchTrain(
			std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* >);
	void handleThresholdScanRun(
			std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* >);
	void handleBeamRun(
			std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* >);
	void handleBeamThresholdScanRun(
			std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* >,
			unsigned);
	void handleCosmicThresholdScanRun(
			std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* >,
			unsigned);

	//Makes a 1D histogram of the number of hits (x axis) 
	//each pixel has integrated over the whole run
	//see findNoisyPixels() below.
	void findHitDistribution();

	//takes the hit pattern integrated over the whole run, 
	//and determines a list of pixels whose number of hits 
	//is greater than 'cut'
	void findNoisyPixels(int cut);

	void makeGlobalTimeStampPlot();

	void computeEfficiencies();

	//writes the ROOT file to disk
	void saveToDisk();

private:
	void populateSensorIds(SubAccessor);
	void makeHistograms(int[]);
	void writeRootFile();
	

	//returns true if this was successful, otherwise false
	bool determineRunType(std::vector<const IlcRunStart*>);

	bool _thresholdScanRun;
	bool _thresholdRun;
	bool _mpsBeamThresholdScan;
	bool _mpsBeam;

	unsigned _runNumber;

	int _sensorIds[5];

	//the index (i.e. 0 - 4 of 5 sensors) that has the PMT readout - default will be set to 100.
	unsigned _pmtDaqIndex;

	int _shaperThreshold[5];
	int _samplerThreshold[5];

	unsigned _nHits;
	unsigned int _tHits;

	unsigned _timeStampsBySensor[5][8192];
	unsigned _timeStampsByShap[5][8192];
	unsigned _timeStampsBySamp[5][8192];

	unsigned _allTimeStamps[8192];
	TH1F* _hTimeStampsBySensor[5];
	TH1F* _hGlobalTimeStamps;
	TH1F* _hSimulTimeStamps;
	
	//These arrays count the number of tracks confirmed and rejected
	unsigned _Tracks[5][2010];
	unsigned _ConfirmedTracks[5][2010];

	unsigned int _bins;
	int _lowEdge;
	int _highEdge;

	unsigned int _nBt;
	unsigned int _specialSensor;

	unsigned int _spillCycleCount;

	TFile* _rootFile;
	TH2I* _hNHitsPerConfig;

	TH2F* _hTimestampsByThreshold[5][4];

	TH1F* _hHitsInRow;
	TH2F* _hHitsInBtByThreshold[5];
	TH2F* _hWordHits[5][4];
	TH2F* _hHits[5][4];
	TProfile* _hHitsByRegion[5][4];
	TH2F* _hGroupsAndRows[5][4];
	TH2F* _hHitPattern[5];
	TH1F* _hThreshEfficacy[5];
	TH1F* _hTimestamps;
	TProfile* _hEffAndThreshShap[5];
	TProfile* _hEffAndThreshSamp[5];
	TProfile* _hEffAndThreshOverall[5];
	TProfile* _hEffAndThreshShapDispT[5];
	TProfile* _hEffAndThreshSampDispT[5];
	TProfile* _hEffAndThreshOverallDispT[5];
	
	//Equivalent to _hEffAndThreshOverall but filled at the end of the run => errors will be different
	TH1F* _hEffIntegratedAllBts[5];

	//histograms for run quality monitoring
	TH1F* _hNHitsInBunchtrain[5];

	TH1F* _hTracks3Vertices;
	TH1F* _hTracks3VerticesInT;
	TH1F* _hTracks4VerticesInT;
	TH2F* _hTracks3VerticesInXY;
	TH2F* _hTracks4VerticesInXY;

	TH2F* _hSoftPhotonsSpecialSensor;
	typedef std::pair<int, int> uv_key;
	typedef std::map<uv_key, TProfile*> thresholdPfxs;
	thresholdPfxs _threshProfs;

	std::vector<const MpsSensor1BunchTrainDatum*> _datums;
	int _configCount;
	typedef std::vector<TObject*> tObjs;
	tObjs myObjs;
	tObjs globals;
	tObjs hitsByRegion;
	tObjs wordsByRegion;
	tObjs groupsAndRows;
	tObjs efficiencies;
	tObjs efficienciesDispT;
	tObjs dqm;
};
//Because we're constrained by an interesting code structure,
//we have to do this filthy thing...
#include "MpsAnalysis.cc"
#endif
