#include <signal.h>

#include <iostream>
#include <sstream>
#include <vector>

#include "UtlArguments.hh"
#include "RcdArena.hh"
#include "RcdCount.hh"
#include "RunReader.hh"
#include "SubAccessor.hh"

#include "MpsAnalysis.hh"

MpsAnalysis::MpsAnalysis(unsigned wombat, int runNumber, int bins, int low,
		int high, int trackingSensor) :
	_thresholdScanRun(false), _thresholdRun(false),
			_mpsBeamThresholdScan(false), _mpsBeam(false), _nBt(0),
			_configCount(0) {
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	_runNumber = runNumber;
	_bins = bins;
	_lowEdge = low;
	_highEdge = high;
	//Assuming we have N boards, which one is considered for special attention in this code
	_specialSensor = trackingSensor;

	//define (unphysical) sensor Ids for now; they'll get populated by populateSensorIds()
	_sensorIds[0] = 95;
	_sensorIds[1] = 96;
	_sensorIds[2] = 97;
	_sensorIds[3] = 98;
	_sensorIds[4] = 99;
	_pmtDaqIndex = 100;
}

MpsAnalysis::~MpsAnalysis() {
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	saveToDisk();
	delete _rootFile;
	std::cout << "~MpsAnalysis(): done. Destructing."<< std::endl;
}

void MpsAnalysis::saveToDisk() {
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	writeRootFile();
	_rootFile->Close();
}

void MpsAnalysis::computeEfficiencies() {
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	double eff(0);
	for(unsigned sensor(0); sensor < 5; sensor++) {
		for(int thresh(0); thresh < _highEdge; thresh++) {
			if(_ConfirmedTracks[sensor][thresh] > 0) {
				eff = static_cast<double>(_ConfirmedTracks[sensor][thresh]) / _Tracks[sensor][thresh] * 100;
				_hEffIntegratedAllBts[sensor]->Fill(thresh, eff);
			}
		}
	}
}

void MpsAnalysis::findNoisyPixels(int cut) {
	std::cout << std::cout << __PRETTY_FUNCTION__ << " with cut = "<< cut
			<< std::endl;
	ofstream highpixels;
	highpixels.open("highpixels.txt");
	unsigned numberOfHighPixels(0);

	for (unsigned sensor(0); sensor < 5; sensor++) {
		highpixels << "Sensor pseudo value: \t"<< sensor << std::endl;

		for (unsigned x(0); x < 168; x++) {
			for (unsigned y(0); y < 168; y++) {
				double hits = _hHitPattern[sensor]->GetBinContent(x, y);
				_hThreshEfficacy[sensor]->Fill(hits);
				if (hits > cut) {
					//naughty pixel! you're going dowwwwwnnn....
					highpixels << x << "\t"<< y << "\t"<< hits << "\n";
					numberOfHighPixels++;
				}
			}
		}
		highpixels << std::endl;
	}
	std::cout << "\tFound "<< numberOfHighPixels << " high pixels"<< std::endl;
	highpixels.close();
}

void MpsAnalysis::findHitDistribution() {
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	for (unsigned sensor(0); sensor < 5; sensor++) {
		for (unsigned x(0); x < 168; x++) {
			for (unsigned y(0); y < 168; y++) {
				double hits = _hHitPattern[sensor]->GetBinContent(x, y);
				_hThreshEfficacy[sensor]->Fill(hits);
			}
		}
	}
}

void MpsAnalysis::makeGlobalTimeStampPlot() {
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	for (unsigned sensor(0); sensor < 5; sensor++) {
		_hGlobalTimeStamps->Add(_hTimeStampsBySensor[sensor]);
	}
}

bool MpsAnalysis::determineRunType(
		std::vector<const IlcRunStart*> runStartHeader) {

	assert(runStartHeader.size()==1);

	for (unsigned i(0); i<runStartHeader.size(); i++) {
		//vs[i]->print(std::cout) << std::endl;
		_runNumber = runStartHeader[i]->runNumber();
		// Check if it is a MAPS threshold scan run
		_thresholdScanRun=(runStartHeader[i]->runType().type()==IlcRunType::mpsThresholdScan);
		//or one in the beam
		_mpsBeamThresholdScan=(runStartHeader[i]->runType().type()==IlcRunType::mpsBeamThresholdScan);
		//or just a threshold run
		_thresholdRun = (runStartHeader[i]->runType().type()==IlcRunType::mpsThreshold);

		_mpsBeam = (runStartHeader[i]->runType().type()==IlcRunType::mpsBeam);
	}
	//if we are able to do the run, return true...
	if (_thresholdScanRun || _thresholdScanRun || _mpsBeamThresholdScan
			|| _thresholdRun || _mpsBeam)
		return true;

	//otherwise we're not set up to handle this
	return false;
}

void MpsAnalysis::populateSensorIds(SubAccessor myAccessor) {
	std::vector<const MpsLocationData<MpsPcb1ConfigurationData>* >
			v(myAccessor.access< MpsLocationData<MpsPcb1ConfigurationData> >());
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	//There are nWrites which preceed a number of reads - we get the
	// sensor Ids from the reads
	unsigned nWrites(0);
	for (unsigned i(0); i<v.size(); i++) {
		if (!v[i]->write()) {
			std::cout<< "\tSensor id for index "<< i << ":\t"
					<< static_cast<unsigned short>(v[i]->location().sensorId()) << std::endl;
			if(v[i]->location().usbDaqMaster())
			{
				//this one is the usbDaqMaster and has the PMTs plugged into it.
				//record this here,
				_pmtDaqIndex = i-nWrites;
				std::cout << "\tUsbDaqMaster is on sensor " << v[i]->location().sensorId() << ", index " << i-nWrites << std::endl;
			}

			_sensorIds[i-nWrites] = v[i]->location().sensorId();
		} else {
			nWrites++;
		}
	}
	std::cout << "Found "<< nWrites
			<< " writes in PCB configuration data, with v.size() = "<< v.size()
			<< std::endl;
}

bool MpsAnalysis::record(const RcdRecord& r) {

	//std::cout<< __PRETTY_FUNCTION__<< std::endl;

	//	if (doPrint(r.recordType())) {
	//		r.RcdHeader::print(std::cout, " ") << std::endl;
	//	}

	SubAccessor accessor(r);

	switch (r.recordType()) {

	case RcdHeader::runStart: {

		// Get list of IlcRunStart subrecords
		// There should only be one
		std::vector<const IlcRunStart*> vs(accessor.access<IlcRunStart>());
		determineRunType(vs);
		break;
	}

	case RcdHeader::runEnd: {
		if (_thresholdScanRun) {
			// Get list of IlcRunEnd subrecords
			// There should only be one
			std::vector<const IlcRunEnd*> ve(accessor.access<IlcRunEnd>());
			assert(ve.size()==1);
			for (unsigned i(0); i<ve.size(); i++) {
				//ve[i]->print(std::cout) << std::endl;
			}
		}
		break;
	}

	case RcdHeader::configurationStart: {
		std::vector<const IlcConfigurationStart*>
				vs(accessor.access<IlcConfigurationStart>());
		assert(vs.size()==1);

		// Get a list of PCB configuration data objects
		std::vector<const MpsLocationData<MpsPcb1ConfigurationData>* >
				v(accessor.access< MpsLocationData<MpsPcb1ConfigurationData> >());

		if (_configCount == 0) {
			//get sensor ids
			populateSensorIds(accessor);
			//make histograms now we know the sensor Ids
			makeHistograms(_sensorIds);
		}
		_configCount++;
		// Get list of IlcConfigurationStart subrecords
		// There should only be one

		
		unsigned nWrites(0);
		for (unsigned i(0); i<v.size(); i++) {
			//the first one is a write, the five remaining are reads
			//v[i]->print(std::cout) << std::endl;
			if (!v[i]->write()) {
				// Get threshold differences
				_shaperThreshold[i - nWrites]=v[i]->data()->region01ThresholdValue();
				_samplerThreshold[i - nWrites]=v[i]->data()->region23ThresholdValue();
			} else {
				nWrites++;
			}
		}

		//a mpsBeamThresholdScan includes an efficiency calculation
		if (_mpsBeamThresholdScan) {
			_thresholdScanRun = true;
		}

		std::vector<const MpsLocationData<MpsUsbDaqConfigurationData>* >
				w(accessor.access<MpsLocationData<MpsUsbDaqConfigurationData> >());

		_spillCycleCount = w[0]->data()->spillCycleCount();

		// Initialise hit counter
		_nHits=0;
		_tHits = 0;
		//_nBt = 0;

		break;

	}

	case RcdHeader::configurationEnd: {
		if (_thresholdScanRun) {

			// Get list of IlcConfigurationEnd subrecords
			// There should only be one
			std::vector<const IlcConfigurationEnd*>
					vs(accessor.access<IlcConfigurationEnd>());

			assert(vs.size()==1);

			for (unsigned i(0); i<vs.size(); i++) {
				//vs[i]->print(std::cout) << std::endl;
			}

		}

		break;
	}

	case RcdHeader::bunchTrain: {
		std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* >
				v(accessor.access< MpsLocationData<MpsSensor1BunchTrainData> >());
		handleBunchTrain(v);
		break;
	}

	default: {
		break;
	}
	};

	return true;
}

void MpsAnalysis::handleBunchTrain(
		std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* > v) {
	if (_nBt%1000== 0) {
		std::cout << "Processing bunch train: "<< _nBt << std::endl;
	}
	if (_mpsBeam) {
		handleBeamRun(v);
	}

	if (_mpsBeamThresholdScan) {
		//is the threshold being varied, or are we at nominal?
		unsigned scannedSensor(0);
		if ((_configCount - 1)%41!= 0) {
			//which sensor is being scanned?
			scannedSensor = static_cast<int>(((_configCount -1)%41- 1)/10);
		}

		handleBeamThresholdScanRun(v, scannedSensor);
	}

	if (_thresholdScanRun) {
		handleThresholdScanRun(v);
	}
	_nBt++;
}

void MpsAnalysis::handleCosmicThresholdScanRun(std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* > v,
		unsigned scannedSensor) {

}

void MpsAnalysis::handleBeamThresholdScanRun(
		std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* > v,
		unsigned scannedSensor) {
	memset(_timeStampsByShap, 0, 5*8192*sizeof(unsigned));
	memset(_timeStampsBySamp, 0, 5*8192*sizeof(unsigned));
	//loop over all the sensors - the first is the PMT, so we ignore it
	for (unsigned i(0); i<v.size(); i++) {
		if (!v[i]->write()) {
			std::vector<MpsSensor1Hit> vh(v[i]->data()->hitVector());
			for (unsigned j(0); j<vh.size(); j++) {
				if (vh[j].pixelX() > 83) {
					//sampler
					_timeStampsBySamp[i][vh[j].timeStamp()]++;
				} else {
					//shaper
					_timeStampsByShap[i][vh[j].timeStamp()]++;
				}
				assert(vh[j].timeStamp() < 8192);

			}
		} else {
			std::cout<< "RcdHeader in bunch train is a write? "<< std::endl;
		}
	}
	//How many tracks there are in this BT
	unsigned nPossibleTracks(0);

	//how many hits are there in the threshold-scanned sensor, when a track in made at a given BX
	unsigned nScannedVeto(0);
	unsigned nScannedVetoShap(0);
	unsigned nScannedVetoSamp(0);

	//define DispT to mean 4096 BX from the examined BX (forward or backward)
	//so that we can have a fake-noise calculation
	//(4906 BXs should not have correlated behaviour with the current BX)
	unsigned nScannedVetoDispT(0);
	unsigned nScannedVetoShapDispT(0);
	unsigned nScannedVetoSampDispT(0);

	unsigned nPossibleTracksNotSpecial(0);

	//now we count hits
	for (unsigned t(0); t < 8192; t++) {
		unsigned hitsInOtherSensorsShap(0);
		unsigned hitsInOtherSensorsSamp(0);
		unsigned hitsInScannedSensorShap(0);
		unsigned hitsInScannedSensorSamp(0);

		//for the de-correlated hit count
		unsigned hitsInScannedSensorShapDispT(0);
		unsigned hitsInScannedSensorSampDispT(0);

		//for soft photon studies
		unsigned hitsInSpecialSensor(0);
		unsigned minHits(1);
		//Now loop over the sensors and count how many hits each sensor has
		for (unsigned sensor(0); sensor < 5; sensor++) {
			_hTimeStampsBySensor[sensor]->Fill(_timeStampsBySensor[sensor][t]);
			if (sensor != scannedSensor) {
				if (_timeStampsBySamp[sensor][t] >= minHits)
					hitsInOtherSensorsSamp++;
				if (_timeStampsByShap[sensor][t] >= minHits)
					hitsInOtherSensorsShap++;
			} else {
				if (_timeStampsBySamp[sensor][t] >= minHits)
					hitsInScannedSensorSamp++;
				if (_timeStampsByShap[sensor][t] >= minHits)
					hitsInScannedSensorShap++;
				//define some displaced time
				unsigned dispT(0);
				if (t >= 4096)
					dispT = t - 4096;
				else
					dispT = t + 4096;

				if (_timeStampsBySamp[sensor][dispT] >= minHits)
					hitsInScannedSensorSampDispT++;
				if (_timeStampsByShap[sensor][dispT] >= minHits)
					hitsInScannedSensorShapDispT++;

			}
			if (sensor == _specialSensor) {
				hitsInSpecialSensor+= _timeStampsBySamp[sensor][t]
						+ _timeStampsByShap[sensor][t];
			} else {
				nPossibleTracksNotSpecial+= _timeStampsBySamp[sensor][t]
						+ _timeStampsByShap[sensor][t];
			}

		}
		//how many vertices are there for to make a possible track
		unsigned nTrackVertex(hitsInOtherSensorsSamp+ hitsInOtherSensorsShap);
		//how many hits are there in the threshold-scanned sensor
		unsigned nScanned(hitsInScannedSensorSamp+ hitsInScannedSensorShap);
		//as above, but at the displaced time
		unsigned nScannedDispT(hitsInScannedSensorSampDispT
				+ hitsInScannedSensorShapDispT);

		if (nTrackVertex > 2) {
			//we assume it's a track
			nPossibleTracks++;
			_hTracks3VerticesInT->Fill(t);
			//new histo - fill with threshold, number of hits in sensor four
			if (scannedSensor == _specialSensor) {
				_hSoftPhotonsSpecialSensor->Fill(
						_shaperThreshold[scannedSensor], hitsInSpecialSensor);
			}
			//does the threshold-scanned sensor confirm a track?
			if (nScanned > 0) {
				nScannedVeto++;
				_hTracks4VerticesInT->Fill(t);
			}
			if (hitsInScannedSensorShap > 0)
				nScannedVetoShap++;
			if (hitsInScannedSensorSamp > 0)
				nScannedVetoSamp++;

			//now for the decorrelated ones
			if (nScannedDispT > 0)
				nScannedVetoDispT++;
			if (hitsInScannedSensorShapDispT > 0)
				nScannedVetoShapDispT++;
			if (hitsInScannedSensorSampDispT > 0)
				nScannedVetoSampDispT++;

			_hTracks3Vertices->Fill(_shaperThreshold[scannedSensor]);

		}

	}
	if (nPossibleTracks > 0) {
		//grind through the efficiency calculations
		//eff = number scanned sensor saw / total possible x 100%
		
		_Tracks[scannedSensor][_shaperThreshold[scannedSensor]] += nPossibleTracks;
		_ConfirmedTracks[scannedSensor][_shaperThreshold[scannedSensor]] += nScannedVeto;

		_hEffAndThreshOverall[scannedSensor]->Fill(
				_shaperThreshold[scannedSensor],
				static_cast<double>(nScannedVeto)/(nPossibleTracks)* 100);

		_hEffAndThreshSamp[scannedSensor]->Fill(
				_shaperThreshold[scannedSensor],
				static_cast<double>(nScannedVetoSamp)/(nPossibleTracks)*100);

		_hEffAndThreshShap[scannedSensor]->Fill(
				_shaperThreshold[scannedSensor],
				static_cast<double>(nScannedVetoShap)/(nPossibleTracks)*100);

		_hEffAndThreshOverallDispT[scannedSensor]->Fill(
				_shaperThreshold[scannedSensor],
				static_cast<double>(nScannedVetoDispT)/(nPossibleTracks)* 100);

		_hEffAndThreshSampDispT[scannedSensor]->Fill(
				_shaperThreshold[scannedSensor],
				static_cast<double>(nScannedVetoSampDispT)/(nPossibleTracks)
						*100);

		_hEffAndThreshShapDispT[scannedSensor]->Fill(
				_shaperThreshold[scannedSensor],
				static_cast<double>(nScannedVetoShapDispT)/(nPossibleTracks)
						*100);

	}

}

void MpsAnalysis::handleBeamRun(
		std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* > v) {

	memset(_timeStampsBySensor, 0, 5*8192*sizeof(unsigned));
	memset(_allTimeStamps, 0, 8192*sizeof(unsigned));
	for (unsigned i(0); i<v.size(); i++) {
		if (!v[i]->write()) {
			std::vector<MpsSensor1Hit> vh(v[i]->data()->hitVector());
			for (unsigned j(0); j<vh.size(); j++) {
				_hHitPattern[i]->Fill(vh[j].pixelX(), vh[j].pixelY());
				assert(vh[j].timeStamp() < 8192);
				_timeStampsBySensor[i][vh[j].timeStamp()]++;
				_allTimeStamps[vh[j].timeStamp()]++;
				//std::cout << vh[j].timeStamp() <<"  :  "<< _timeStampsBySensor[i][vh[j].timeStamp()]<<std::endl;
			}
		}
	}
	for (unsigned sensor(0); sensor < 5; sensor++) {
		for (unsigned t(0); t < 8192; t++) {
			_hTimeStampsBySensor[sensor]->Fill(_timeStampsBySensor[sensor][t]);
			if (sensor ==0)
				_hSimulTimeStamps->Fill(_allTimeStamps[t]);
		}
	}

}

void MpsAnalysis::handleThresholdScanRun(
		std::vector<const MpsLocationData<MpsSensor1BunchTrainData>* > v) {
	//Iterate over each sensor labelled by 'i'
	for (unsigned i(0); i<v.size(); i++) {
		if (!v[i]->write()) {

			_hHitsInBtByThreshold[i]->Fill(_shaperThreshold[i], v[i]->data()->numberOfRegionHits(0)+ v[i]->data()->numberOfRegionHits(1));

			for (unsigned theRegion(0); theRegion < 4; theRegion++) {
				//Need to loop over word hits
				unsigned allHitsA(0);
				unsigned allHitsB(0);
				unsigned wordHitsA(0);
				unsigned wordHitsB(0);
				const MpsSensor1BunchTrainDatum* rDatum =v[i]->data()->regionData(theRegion);

				for (unsigned j(0); j < v[i]->data()->numberOfRegionHits(theRegion); j++) {
					for (unsigned c(0); c < 6; c++) {
						if (rDatum[j].channel(c) != 0&& rDatum[j].row()<84) {
							allHitsA++;
						} else if (rDatum[j].channel(c) != 0&& rDatum[j].row()
								> 83) {
							allHitsB++;
						}
					}
					if (rDatum[j].row() < 84)
						wordHitsA++;
					if (rDatum[j].row() > 83)
						wordHitsB++;
					_hGroupsAndRows[i][theRegion]->Fill(rDatum[j].group(),
							rDatum[j].row());
				}
				unsigned thresh(_shaperThreshold[i]);
				if (theRegion == 2|| theRegion == 3) {
					thresh = _samplerThreshold[i];
				}
				_hHits[i][theRegion]->Fill(thresh, allHitsA+ allHitsB);
				_hHitsByRegion[i][theRegion]->Fill(thresh, allHitsA+ allHitsB);
				_hWordHits[i][theRegion]->Fill(thresh, wordHitsA+ wordHitsB);

			}

			std::vector<MpsSensor1Hit> vh(v[i]->data()->hitVector());

			for (unsigned j(0); j<vh.size(); j++) {
				_hHitPattern[i]->Fill(vh[j].pixelX(), vh[j].pixelY());
				unsigned thresh(_shaperThreshold[i]);
				if (vh[j].pixelX() > 83) {
					thresh = _samplerThreshold[i];
				}
				_hTimestampsByThreshold[i][static_cast<int>(vh[j].pixelX()/42)]->Fill(thresh, vh[j].timeStamp());
				assert(vh[j].timeStamp() < 8192);
			}
		} else {
			std::cout<< "RcdHeader in bunch train is a write? "<< std::endl;
		}
	}
}

void MpsAnalysis::makeHistograms(int sensorIds[]) {
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	std::string rootFileTitle(("MpsAnalysis_" + IntToStr(_runNumber)+ ".root").c_str());
	_rootFile = new TFile(rootFileTitle.c_str(),"RECREATE");
	_hNHitsPerConfig
			= new TH2I("hNHitsPerConfig", "Number of hits by threshold", _bins, _lowEdge, _highEdge, 30, 0, 30);
	std::cout << "\tInitialising histograms..."<< std::endl;
	std::cout << "\tbins: "<< _bins << ", lowEdge: "<< _lowEdge
			<< ", _highEdge: "<< _highEdge << std::endl;

	_hHitsInRow
			= new TH1F("hHitsInRow","Hits in a row over all BTs, for one thresh;Pixel",42,0,42);

	_hGlobalTimeStamps = new TH1F("hGlobalTimeStamps", "hGlobalTimeStamps", 40, 0, 40);
	globals.push_back(_hGlobalTimeStamps);

	_hSimulTimeStamps = new TH1F("hSimulTimeStamps", "hSimulTimeStamps", 40, 0, 40);
	globals.push_back(_hSimulTimeStamps);

	memset(_timeStampsBySensor, 0, 5*8192*sizeof(unsigned));
	memset(_allTimeStamps, 0, 8192*sizeof(unsigned));
	memset(_Tracks, 0, 5*2010*sizeof(unsigned));
	memset(_ConfirmedTracks, 0, 5*2010*sizeof(unsigned));

	_hSoftPhotonsSpecialSensor
			= new TH2F("hSoftPhotonsSpecialSensor", "Hits in non-tracking front sensor", _bins, _lowEdge, _highEdge, 100, 0, 100);
	efficiencies.push_back(_hSoftPhotonsSpecialSensor);

	_hTracks3Vertices
			= new TH1F("hTracks3Vertices", "Number of tracks with 3 vertices", _bins, _lowEdge, _highEdge);
	efficiencies.push_back(_hTracks3Vertices);
	
	_hTracks3VerticesInT = new TH1F("hTracks3VerticesInT", "Track (3VX) formation in time", 8192/32, 0, 8192);
	efficiencies.push_back(_hTracks3VerticesInT);
	_hTracks4VerticesInT = new TH1F("hTracks4VerticesInT", "Track (4VX) formation in time", 8192/32, 0, 8192);
	efficiencies.push_back(_hTracks4VerticesInT);

	//define a hit pattern for each of 5 possible sensors
	for (unsigned s(0); s < 5; s++) {
		_hEffIntegratedAllBts[s] = new TH1F(("hEffIntegratedAllBts" + IntToStr(sensorIds[s])).c_str(), ("Efficiency integrated over all BTs, sensor " + IntToStr(sensorIds[s])).c_str(), _bins, _lowEdge, _highEdge);
		efficiencies.push_back(_hEffIntegratedAllBts[s]);

		_hTracks3VerticesInXY[s] = new TH2F(("hTracks3VerticesInXY" + IntToStr(sensorIds[s])).c_str(), ("hTracks3VerticesInXY" + IntToStr(sensorIds[s])).c_str(),168, 0, 168, 168, 0, 168);
		efficiencies.push_back(_hTracks3VerticesInXY[s]);
		
		_hTracks4VerticesInXY[s] = new TH2F(("hTracks4VerticesInXY" + IntToStr(sensorIds[s])).c_str(), ("hTracks4VerticesInXY" + IntToStr(sensorIds[s])).c_str(),168, 0, 168, 168, 0, 168);
		efficiencies.push_back(_hTracks4VerticesInXY[s]);

		_hHitPattern[s]
				= new TH2F(("hHitPattern" + IntToStr(sensorIds[s])).c_str(), "Integrated hit pattern;Pixel X;Pixel Y",168,0,168,168,0,168);
		globals.push_back(_hHitPattern[s]);

		_hThreshEfficacy[s]
				= new TH1F(("hThreshEfficacy"+ IntToStr(sensorIds[s])).c_str(), "Threshold Efficacy;Hit count per pixel", 20000, 0, 20000);
		globals.push_back(_hThreshEfficacy[s]);

		_hHitsInBtByThreshold[s]
				= new TH2F(("hHitsInBtByThreshold" + IntToStr(sensorIds[s])).c_str(), "Hits in BT by threshold;Threshold; NHits", _bins, _lowEdge, _highEdge, 100, 0, 100);
		globals.push_back(_hHitsInBtByThreshold[s]);

		_hTimeStampsBySensor[s] = new TH1F(("hTimeStampsBySensor"+ IntToStr(_sensorIds[s])).c_str(), ("hTimeStampsBySensor"+ IntToStr(s)).c_str(), 40, 0, 40);
		globals.push_back(_hTimeStampsBySensor[s]);

		_hEffAndThreshSamp[s] = new TProfile(("hEffAndThreshSamp" + IntToStr(sensorIds[s])).c_str(), ("hEffAndThreshSamp" + IntToStr(sensorIds[s])).c_str(), _bins, _lowEdge, _highEdge);
		_hEffAndThreshShap[s] = new TProfile(("hEffAndThreshShap" + IntToStr(sensorIds[s])).c_str(), ("hEffAndThreshShap" + IntToStr(sensorIds[s])).c_str(), _bins, _lowEdge, _highEdge);
		_hEffAndThreshOverall[s]= new TProfile(("pEffAndThreshOverall" + IntToStr(sensorIds[s])).c_str(), ("Overall efficiency, sensor " + IntToStr(sensorIds[s])).c_str(), _bins, _lowEdge, _highEdge);
		efficiencies.push_back(_hEffAndThreshShap[s]);
		efficiencies.push_back(_hEffAndThreshSamp[s]);
		efficiencies.push_back(_hEffAndThreshOverall[s]);

		_hEffAndThreshSampDispT[s] = new TProfile(("hEffAndThreshSampDispT" + IntToStr(sensorIds[s])).c_str(), ("hEffAndThreshSampDispT" + IntToStr(sensorIds[s])).c_str(), _bins, _lowEdge, _highEdge);
		_hEffAndThreshShapDispT[s] = new TProfile(("hEffAndThreshShapDispT" + IntToStr(sensorIds[s])).c_str(), ("hEffAndThreshShapDispT" + IntToStr(sensorIds[s])).c_str(), _bins, _lowEdge, _highEdge);
		_hEffAndThreshOverallDispT[s]= new TProfile(("pEffAndThreshOverallDispT" + IntToStr(sensorIds[s])).c_str(), ("Overall efficiency, sensor " + IntToStr(sensorIds[s])).c_str(), _bins, _lowEdge, _highEdge);
		efficienciesDispT.push_back(_hEffAndThreshShapDispT[s]);
		efficienciesDispT.push_back(_hEffAndThreshSampDispT[s]);
		efficienciesDispT.push_back(_hEffAndThreshOverallDispT[s]);

		_hNHitsInBunchtrain[s] = new TH1F(("hNHitsInBunchtrain" + IntToStr(sensorIds[s])).c_str(), ("hNHitsInBunchTrain" + IntToStr(sensorIds[s])).c_str(), 1000000, 0, 1000000);
		dqm.push_back(_hNHitsInBunchtrain[s]);

		for (unsigned r(0); r < 4; r++) {

			_hHits[s][r] = new TH2F(("_hHitsRegionS" + IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), ("hHitsRegionS" + IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), _bins, _lowEdge, _highEdge, 400, 0, 400);
			hitsByRegion.push_back(_hHits[s][r]);

			_hWordHits[s][r] = new TH2F(("_hWordHitsRegionS" + IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), ("hWordHitsRegionS" + IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), _bins, _lowEdge, _highEdge, 400, 0, 4000);
			wordsByRegion.push_back(_hWordHits[s][r]);

			_hGroupsAndRows[s][r] = new TH2F(("_hGroupsAndRowsS" + IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), ("hGroupsAndRowsS" + IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), 7, -0.5, 6.5, 168, 0, 168);
			groupsAndRows.push_back(_hGroupsAndRows[s][r]);

			_hHitsByRegion[s][r] = new TProfile(("_hHitsByRegionS" + IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), ("_hHitsByRegionS" + IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), _bins, _lowEdge, _highEdge);
			hitsByRegion.push_back(_hHitsByRegion[s][r]);

			_hTimestampsByThreshold[s][r] = new TH2F(("hTimestampsByThreshS"+ IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), ("hTimestampsByThreshS"+ IntToStr(sensorIds[s]) + "R" + IntToStr(r)).c_str(), _bins, _lowEdge, _highEdge, 200, 0, 200);
			hitsByRegion.push_back(_hTimestampsByThreshold[s][r]);
		}
	}
	std::cout << "\tFinishing up in " << __PRETTY_FUNCTION__ << std::endl;
}

void MpsAnalysis::writeRootFile() {
	std::cout << __PRETTY_FUNCTION__ << std::endl;
	_rootFile->cd("/");
	for (tObjs::iterator it = globals.begin(); it != globals.end(); it++) {
		(*it)->Write();
	}

	_rootFile->cd("/");
	_rootFile->mkdir("hitsByRegion")->cd();
	for (tObjs::iterator it = hitsByRegion.begin(); it!= hitsByRegion.end(); it++) {
		(*it)->Write();
	}

	_rootFile->cd("/");
	_rootFile->mkdir("wordsByRegion")->cd();
	for (tObjs::iterator it = wordsByRegion.begin(); it!= wordsByRegion.end(); it++) {
		(*it)->Write();
	}

	_rootFile->cd("/");
	_rootFile->mkdir("groupsAndRows")->cd();
	for (tObjs::iterator it = groupsAndRows.begin(); it!= groupsAndRows.end(); it++) {
		(*it)->Write();
	}

	_rootFile->cd("/");
	_rootFile->mkdir("efficiencies")->cd();
	for (tObjs::iterator it = efficiencies.begin(); it!= efficiencies.end(); it++) {
		(*it)->Write();
	}

	_rootFile->mkdir("dispTeff")->cd();
	for (tObjs::iterator it = efficienciesDispT.begin(); it
			!= efficienciesDispT.end(); it++) {
		(*it)->Write();
	}

	_rootFile->cd("/");
	_rootFile->mkdir("dqm")->cd();
	for (tObjs::iterator it = dqm.begin(); it!= dqm.end(); it++) {
		(*it)->Write();
	}
}
